#
#  ProTRG.pyw
#  ProTRG
#
#  Created by poiuy_qwert on 29/09/09.
#

# Done:
# * Fixed LeaderboardComputerPlayers (changed parameter from SETSWITCHSTATE to SETOTHERSTATE)
# * Fixed the default trigger file from "triggers.txt" to "triggers.protrg"
# * Fixed the convenience global ACTIONS
# * Fixed some TUNIT vs UNIT inconsistancies, and the definitions for conditions/actions with the PERCENT parameter type
# * Fixed multilines and slashes in string conversion
# * Fixed potential problem with script filenames starting with "protrg" (without quotes)
# * Fixed switches to be numbered the same as SCMDraft switches (1-256 instead of 0-255)
# * Fixed many unit names
# * Added support for compiling straight to .trg's, including GOT compatible .trg's (Note: StarEdit has problems loading recycled strings)
# * Added an optional mechanism to have stronger typed values for parameters (Explained more in the Coding section)
# * Added an optional mechanism to have stronger typed values for parameters (Explained more in the Coding section)
# * Added support for Extended Unit Deaths (EUD's) using the new stronger type mechanism (using EXTENDEDUNIT type)
# * Added DefaultString, which is set to an empty string (""), and all default AI scripts as Globals
# * Added @iclude preprocessor to iclude ProTRG files in others
# * Added better command line support
# * Added better support for Properties
# * Added the keyword argument "disabled" for conditions/actions
# * Added the possibility to collapse code in Notepad++ using #> and #< for the start and end
# * Added a list of the Globals and their values to reference.txt
# * LOCATION id's are now 0 to 254 instead of 1 to 255 to conform with map editors default location names
# * If an error happens in your script it will now print all the current triggers in the Trigger list (if any) after printing the error traceback
# * All conditions and actions now inherit from the Condition and Action classes instead of directly from Raw

VERSION = (1,1)
LONG_VERSION = 'v%s.%s' % VERSION

from Globals import *
import sys, os.path, traceback, random, time, math, re, optparse, struct
sys.stderr = open('stdeo.txt','w')

if hasattr(sys, 'frozen'):
	PROTRG_PATH,protrg = os.path.split(unicode(sys.executable, sys.getfilesystemencoding()))
else:
	PROTRG_PATH = sys.path[0]
	protrg = os.path.basename(unicode(__file__, sys.getfilesystemencoding()))

class Triggers:
	def __init__(self):
		self.triggerList = []
		self.lastTrigger = None
		self.properties = []
		# Only used for .trg's:
		self.strings = []
		self.added = []
		self.next = []
		self.trig = 0 # 1 = .trg, 2 = GOT .trg

	def add(self, trigger):
		self.triggerList.append(trigger)
		self.lastTrigger = trigger

TRIGGERS = Triggers()

class ProTRGError(Exception):
	def __init__(self, t, d):
		self.t = t
		self.d = d

	def __str__(self):
		return "%s error: %s" % (self.t,self.d)

def isnum(val):
	if isinstance(val, ProTRGTYPE):
		val = val.value
	return isinstance(val, (int,long))
def isstr(val):
	if isinstance(val, ProTRGTYPE):
		val = val.value
	return isinstance(val, (str,unicode))

class ProTRGTYPE:
	def __init__(self, val):
		self.value = val
	def new(self, val):
		return self.__class__(val)
	def __repr__(self):
		return '%s(%s)' % (self.__class__.__name__, repr(self.value))
	def __str__(self):
		return str(self.value)
	# General emulation methods
	def __cmp__(self, other):
		if isnum(self.value):
			return self.value - other
		elif self.value == other:
			return 0
		else:
			return (self.value < other)*-2+1
	def __hash__(self):
		return hash(self.value)
	def __nonzero__(self):
		return not not self.value
	# Numerical emulation methods
	def __add__(self, other):
		try:
			val = self.value + other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __sub__(self, other):
		try:
			val = self.value - other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __mul__(self, other):
		try:
			val = self.value * other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __floordiv__(self, other):
		try:
			val = self.value // other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __mod__(self, other):
		try:
			val = self.value % other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __divmod__(self, other):
		try:
			val = (self.value // other,self.value % other)
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return (self.new(val[0]),self.new(val[1]))
	def __pow__(self, other, modulo=0):
		try:
			val = pow(self.value,other,modulo)
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __lshift__(self, other):
		try:
			val = self.value << other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rshift__(self, other):
		try:
			val = self.value >> other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __and__(self, other):
		try:
			val = self.value & other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __xor__(self, other):
		try:
			val = self.value ^ other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __or__(self, other):
		try:
			val = self.value | other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __div__(self, other):
		try:
			val = self.value / other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __truediv__(self, other):
		try:
			val = self.value / other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __radd__(self, other):
		try:
			val = other + self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rsub__(self, other):
		try:
			val = other - self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rmul__(self, other):
		try:
			val = other * self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rdiv__(self, other):
		try:
			val = other / self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rtruediv__(self, other):
		try:
			val = other / self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rfloordiv__(self, other):
		try:
			val = other // self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rmod__(self, other):
		try:
			val = other % self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rdivmod__(self, other):
		try:
			val = (other // self.value,other & self.value)
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return (self.new(val[0]),self.new(val[1]))
	def __rpow__(self, other, modulo=0):
		try:
			val = pow(other,self.value,modulo)
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rlshift__(self, other):
		try:
			val = other << self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rrshift__(self, other):
		try:
			val = other >> self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rand__(self, other):
		try:
			val = other & self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __rxor__(self, other):
		try:
			val = other ^ self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __ror__(self, other):
		try:
			val = other | self.value
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __iadd__(self, other):
		try:
			val = self.value + other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __isub__(self, other):
		try:
			val = self.value - other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __imul__(self, other):
		try:
			val = self.value * other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self.new(val)
	def __idiv__(self, other):
		try:
			val = self.value / other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __itruediv__(self, other):
		try:
			val = self.value / other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __ifloordiv__(self, other):
		try:
			val = self.value // other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __imod__(self, other):
		try:
			val = self.value % other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __ipow__(self, other, modulo=0):
		try:
			val = pow(self.value,other,modulo)
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __rlshift__(self, other):
		try:
			val = self.value << other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __irshift__(self, other):
		try:
			val = self.value >> other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __iand__(self, other):
		try:
			val = self.value & other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __ixor__(self, other):
		try:
			val = self.value ^ other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		self.value = val
		return self
	def __ior__(self, other):
		try:
			val = self.value | other
		except Exception, e:
			raise ProTRGError('ProTRGTYPE',' '.join(e.args))
		return self
	def __complex__(self):
		if isnum(self.value):
			return complex(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an int so can't be converted to complex" % self.__class__.__name__)
	def __int__(self):
		if isnum(self.value):
			return int(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an int so can't be converted to an int" % self.__class__.__name__)
	def __long__(self):
		if isnum(self.value):
			return long(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an integer so can't be converted to a long" % self.__class__.__name__)
	def __float__(self):
		if isnum(self.value):
			return float(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an integer so can't be converted to a float" % self.__class__.__name__)
	def __oct__(self):
		if isnum(self.value):
			return oct(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an integer so can't be converted to an oct string" % self.__class__.__name__)
	def __hex__(self):
		if isnum(self.value):
			return hex(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an integer so can't be converted to an hex string" % self.__class__.__name__)
	def	__index__(self):
		if isnum(self.value):
			return int(self.value)
		raise ProTRGError('ProTRGTYPE',"The %s is not holding an integer so can't be converted to an index" % self.__class__.__name__)
	def __coerce__(self, other):
		try:
			if isnum(self.value):
				return (self.value, int(other))
			else:
				return (self.value, str(other))
		except:
			pass
		return None

class BYTE(ProTRGTYPE):
	"""Must be a number between 0 and 255"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, BYTE)) or not isnum(val) or val < 0 or val > 255:
			raise ProTRGError('BYTE parameter','%s (got "%s")' % (BYTE.__doc__,repr(val)))
		return val
	def decompile(val):
		return str(val)
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class SHORT(ProTRGTYPE):
	"""Must be a number between 0 and 65535"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, (BYTE,SHORT))) or not isnum(val) or val < 0 or val > 65535:
			raise ProTRGError('SHORT parameter','%s (got "%s")' % (SHORT.__doc__,repr(val)))
		return val
	def decompile(val):
		return str(val)
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class LONG(ProTRGTYPE):
	"""Must be a number lower than 4294967295"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, (BYTE,SHORT,LONG))) or not isnum(val) or val > 4294967295:
			raise ProTRGError('LONG parameter','%s (got "%s")' % (LONG.__doc__,repr(val)))
		return val
	def decompile(val):
		return str(val)
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class PLAYER(ProTRGTYPE):
	"""Must be a player number from 1 to 12, a force name string, or one of:
	Player1-Player12 = 1-12
	CurrentPlayer = 13
	Foes = 14
	Allies = 15
	NeutralPlayers = 16
	AllPlayers = 17
	Force1 = 18
	Force2 = 19
	Force3 = 20
	Force4 = 21
	NonAlliedVictoryPlayers = 26"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, PLAYER)) or ((isnum(val) and (val < 1 or (val > 21 and val != 26))) and not isstr(val)):
			raise ProTRGError('PLAYER parameter','%s\n(got "%s")' % (PLAYER.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			raise ProTRGError('PLAYER paramater','You cannot use custom force name strings when compiling to a .trg (use the globals Force1, Force2, Force3, and Force4 instead)')
		if TRIGGERS.trig and val < 13:
			val -= 1
		return val
	vals = {13:"Current Player",14:"Foes",15:"Allies",16:"Neutral Players",17:"All Players",26:"Non Allied Victory Players"}
	def decompile(val):
		if isstr(val):
			return '"%s"' % val
		elif val < 13:
			return '"Player %i"' % val
		elif val > 17 and val < 22:
			return '"Force %i"' % (val - 17)
		return '"%s"' % PLAYER.vals[val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class COMPARISON(ProTRGTYPE):
	"""Must be one of the three comparison types:
	AtLeast = 0
	AtMost = 1
	Exactly = 10"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, PLAYER)) or  not isnum(val) or not val in [0,1,10]:
			raise ProTRGError('COMPARISON parameter','%s\n(got "%s")' % (COMPARISON.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["At least","At most","Exactly"][val % 8]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class UNIT(ProTRGTYPE):
	"""Must be a unit name string or a unit id number between 0 and 227"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, UNIT)) or (isnum(val) and (val < 0 or val > 227)):
			raise ProTRGError('UNIT parameter','%s (got "%s")' % (UNIT.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			val = str(val)
			if not val in ALL_UNITS:
				raise ProTRGError('UNIT paramater','You cannot use custom unit name strings when compiling to a .trg (use the unit name globals, the default starcraft unit names, or a unit ID between 0 and 227)')
			return ALL_UNITS.index(val)
		return val
	def decompile(val):
		if isnum(val):
			return '"%s"' % ALL_UNITS[val]
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class EXTENDEDUNIT(ProTRGTYPE):
	"""Must be a unit name string or a unit id number between 0 and 4294967295 (Note: If you pass a UNIT type, the unit id can only be between 0 and 227)"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, (UNIT,EXTENDEDUNIT))) or (isnum(val) and (val < 0 or val > (4294967295,227)[isinstance(val, UNIT)])):
			raise ProTRGError('EXTENDEDUNIT parameter','%s (got "%s")' % (UNIT.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			val = str(val)
			if not val in ALL_UNITS:
				raise ProTRGError('EXTENDEDUNIT paramater','You cannot use custom unit name strings when compiling to a .trg (use the unit name globals, the default starcraft unit names, or a unit ID between 0 and 4294967295)')
			return ALL_UNITS.index(val)
		return val
	def decompile(val):
		if isnum(val):
			if val > 227:
				return 'ID:%s' % val
			return '"%s"' % ALL_UNITS[val]
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class TUNIT(ProTRGTYPE):
	"""Must be a unit name string, a unit id number (0-227), or one of the four unit types:
	NoUnit = "None"
	AnyUnit = "Any Unit"
	Men = "Men"
	Buildings = "Buildings"
	Factories = "Factories\""""
	unit_types = [NoUnit,AnyUnit,Men,Buildings,Factories]
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, (UNIT,TUNIT))) or ((isnum(val) and (val < 0 or val > 227)) and not isstr(val)):
			raise ProTRGError('TUNIT parameter','%s\n(got "%s")' % (TUNIT.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			val = str(val)
			if val in TUNIT.unit_types:
				return 228 + TUNIT.unit_types.index(val)
			if not val in ALL_UNITS:
				raise ProTRGError('TUNIT paramater','You cannot use custom unit name strings when compiling to a .trg (use the unit name globals, the default starcraft unit names, or a unit ID between 0 and 227)')
			return ALL_UNITS.index(val)
		return val
	def decompile(val):
		if isnum(val):
			return '"%s"' % ALL_UNITS[val]
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

location_re = re.compile("^\s*[Ll]ocation\s*(\d{1,3})\s*$")
class LOCATION(ProTRGTYPE):
	"""Must be location name string, or a number from 0 to 254"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, LOCATION)) or ((isnum(val) and val < 0 or val > 254) and not isstr(val)):
			raise ProTRGError('LOCATION parameter','%s (got "%s")' % (LOCATION.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			if str(val) == Anywhere:
				return 64
			m = location_re.match(str(val))
			if m:
				try:
					val = int(m.group(1))
				except:
					val = -1
				if val > -1 and val < 255:
					return val+1
			raise ProTRGError('LOCATION parameter','You cannot use custom location name strings when compiling to a .trg (use a location id from 0 to 254 or a string in the form "location #" where # is a valid location id)')
		if isnum(val):
			val += 1
		return val
	def decompile(val):
		if isnum(val):
			return '"Location %s"' % (val-1)
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class RESOURCE(ProTRGTYPE):
	"""Must be one of the three resource types:
	Ore = 0
	Gas = 1
	OreAndGas = 2"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, RESOURCE)) or not isnum(val) or val < 0 or val > 2:
			raise ProTRGError('RESOURCE parameter','%s\n(got "%s")' % (RESOURCE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ['ore','gas','ore and gas'][val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class SCORE(ProTRGTYPE):
	"""Must be one of 8 score type:
	Total = 0
	Units = 1
	BuildingScore = 2
	UnitsAndBuildings = 3
	Kills = 4
	Razings = 5
	KillsAndRazings = 6
	Custom = 7"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, SCORE)) or not isnum(val) or val < 0 or val > 7:
			raise ProTRGError('SCORE parameter','%s\n(got "%s")' % (SCORE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ['Total','Units','Buildings','Units and buildings','Kills','Razings','Kills and razings','Custom'][val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

switch_re = re.compile("^\s*[Ss]witch\s*(\d{1,3})\s*$")
class SWITCH(ProTRGTYPE):
	"""Must be a switch name string or a switch id number between 1 and 256"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, SWITCH)) or ((isnum(val) and (val < 1 or val > 256)) and not isstr(val)):
			raise ProTRGError('SWITCH parameter','%s (got "%s")' % (SWITCH.__doc__,repr(val)))
		if TRIGGERS.trig and isstr(val):
			m = location_re.match(str(val))
			if m:
				try:
					val = int(m.group(0))
				except:
					val = 0
				if val > 0 and val < 257:
					return val-1
			raise ProTRGError('SWITCH parameter','You cannot use custom switch name strings when compiling to a .trg (use a switch id from 0 to 255 or a string in the form "switch #" where # is a location id)')
		return val-1
	def decompile(val):
		if isnum(val):
			return '"Switch%s"' % (val+1)
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class STATE(ProTRGTYPE):
	"""Must be one of two states:
	IsSet = 2
	IsCleared = 3"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, STATE)) or not isnum(val) or (val != 2 and val != 3):
			raise ProTRGError('STATE parameter','%s\n(got "%s")' % (STATE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["set","not set"][val-2]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class SUBTITLE(ProTRGTYPE):
	"""Must be one of two settings:
	SubtitleDisplay = 0
	AlwaysDisplay = 1"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, SUBTITLE)) or not isnum(val) or not val in [0,1]:
			raise ProTRGError('SUBTITLE parameter','%s\n(got "%s")' % (SUBTITLE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["Don't Always Display","Always Display"][val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class PROPERTY(ProTRGTYPE):
	"""Must be a property id number in the range 0 to 63 or a Property instance"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, PROPERTY)) or ((not isnum(val) or val < 0 or val > 63) and not isinstance(val, Property)):
			raise ProTRGError('PROPERTY parameter','%s (got "%s")' % (PROPERTY.__doc__,repr(val)))
		if TRIGGERS.trig == 1 and isnum(val):
			raise ProTRGError('PROPERTY parameter','You cannot use property id numbers when compiling to a .trg (you must create and use Property instances)')
		if isinstance(val, Property):
			val = TRIGGERS.properties.index(val)
		return val
	def decompile(val):
		return str(val)
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class STRING(ProTRGTYPE):
	"""Must be a string or the keyword:
	DefaultString = """""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, STRING)) or not isstr(val):
			raise ProTRGError('STRING parameter','%s (got "%s")' % (STRING.__doc__,repr(val)))
		if TRIGGERS.trig:
			if TRIGGERS.trig == 2:
				return 0
			else:
				val = str(val)
				if not val:
					return 0
				if len(val) > 2047:
					raise ProTRGError('STRING parameter','Strings cannot exceed 2047 characters when compiling to a .trg (got string of length "%s")' % len(val))
				if not val in TRIGGERS.strings:
					TRIGGERS.strings.append(val)
				return TRIGGERS.strings.index(val)+1
		return val
	def decompile(val):
		return '"%s"' % val.replace('\r\n','<13><10>').replace('\r','<13>').replace('\n','<10>')
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class SETSWITCHSTATE(ProTRGTYPE):
	"""Must be one of four state modifiers:
	Set = 4
	Cleared = 5
	Toggle = 6
	Randomize = 11"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, SETSWITCHSTATE)) or not isnum(val) or not val in [4,5,6,11]:
			raise ProTRGError('SETSWITCHSTATE parameter','%s\n(got "%s")' % (SETSWITCHSTATE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ['set','clear','toggle','randomize'][val % 4]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class SETOTHERSTATE(ProTRGTYPE):
	"""Must be one of three keywords:
	Disabled = 4
	Enabled = 5
	Toggle = 6"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, SETOTHERSTATE)) or not isnum(val) or val < 4 or val > 6:
			raise ProTRGError('SETOTHERSTATE parameter','%s\n(got "%s")' % (SETOTHERSTATE.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["disabled","enabled","toggle"][val-4]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class MODIFIER(ProTRGTYPE):
	"""Must be one of three value modifiers:
	SetTo = 7
	Add = 8
	Subtract = 9"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, MODIFIER)) or not isnum(val) or not val in [7,8,9]:
			raise ProTRGError('MODIFIER parameter','%s\n(got "%s")' % (MODIFIER.__doc__,repr(val)))
		return val
	def decompile(val):
		return ['Set To','Add','Subtract'][val - 7]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class AISCRIPT(ProTRGTYPE):
	"""Must be a 4 letter AI Script ID string"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, AISCRIPT)) or not isstr(val) or len(val) != 4:
			raise ProTRGError('AISCRIPT parameter','%s (got "%s")' % (AISCRIPT.__doc__,repr(val)))
		if TRIGGERS.trig:
			return struct.unpack('<L', str(val))[0]
		return val
	def decompile(val):
		return '"%s"' % val
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class QUANTITY(ProTRGTYPE):
	"""Must be a quantity number from 1 to 255, or for all units use the keyword/value:
	All = 0"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, QUANTITY)) or not isnum(val) or val < 0 or val > 255:
			raise ProTRGError('QUANTITY parameter','%s\n(got "%s")' % (QUANTITY.__doc__,repr(val)))
		return val
	def decompile(val):
		if val:
			return str(val)
		return "All"
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class ORDER(ProTRGTYPE):
	"""Must be one of the three orders:
	Move = 0
	Patrol = 1
	Attack = 2"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, ORDER)) or not isnum(val) or not val in [0,1,2]:
			raise ProTRGError('ORDER parameter','%s\n(got "%s")' % (ORDER.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["move","patrol","attack"][val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class PERCENT(ProTRGTYPE):
	"""Must be a percent number in the range 0 to 100"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, PERCENT)) or not isnum(val) or val < 0 or val > 100:
			raise ProTRGError('PERCENT parameter','%s (got "%s")' % (PERCENT.__doc__,repr(val)))
		return val
	def decompile(val):
		return str(val)
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

class ALLYSTATUS(ProTRGTYPE):
	"""Must be one of three ally status':
	Enemy = 0
	Ally = 1
	AlliedVictory = 2"""
	def compile(val):
		if (isinstance(val, ProTRGTYPE) and not isinstance(val, ALLYSTATUS)) or not isnum(val) or not val in [0,1,2]:
			raise ProTRGError('ALLYSTATUS parameter','%s\n(got "%s")' % (ALLYSTATUS.__doc__,repr(val)))
		return val
	def decompile(val):
		return ["Enemy","Ally","Allied Victory"][val]
	compile = staticmethod(compile)
	decompile = staticmethod(decompile)

# Trigger and Raw and Property class'
class Trigger:
	def __init__(self, *players, **options):
		if not players:
			raise ProTRGError('Trigger','Triggers need at least one player assigned to them.')
		self.players = []
		self.conditions = []
		self.actions = []
		for p in players:
			c = PLAYER.compile(p)
			if not c in self.players:
				self.players.append(c)
		add = options.get('add', True)
		if 'add' in options:
			del options['add']
		if options:
			raise
		if add:
			TRIGGERS.add(self)
		else:
			TRIGGERS.lastTrigger = self

	def __str__(self):
		players = ",".join([PLAYER.decompile(p) for p in self.players])
		conditions = "\n".join([str(c) for c in self.conditions])
		if conditions:
			conditions += '\n'
		actions = "\n".join([str(a) for a in self.actions])
		if actions:
			actions += '\n'
		return "Trigger(%s) {\nConditions:\n%s\nActions:\n%s}\n" % (players,conditions,actions)

	def TRIG(self):
		s = ''
		for c in self.conditions:
			s += c.TRIG()
		s += '\x00' * 20 * (16-len(self.conditions))
		for a in self.actions:
			s += a.TRIG()
		s += '\x00' * 32 * (64-len(self.actions))
		p = [0] * 28
		for n in self.players:
			p[n] = 1
		#print p
		return s + struct.pack('4x28B', *p)

	def add(self, raw):
		if not isinstance(raw, Raw):
			raise
		a = [self.conditions,self.actions][raw.action]
		if len(a) == [16,64][raw.action]:
			t = ['Condition','Action'][raw.action]
			raise ProTRGError('%s amount' % t, 'You are attempting to add a%s %s when you are already at the maximum for the trigger (maximum of %s)' % (['','n'][raw.action], t.lower(), [16,64][raw.action]))
		a.append(raw)

	def addTrigger(self):
		TRIGGERS.triggerList.append(self)

class Raw:
	action = None
	definition = None
	disabled = False
	id = None
	name = None

	def __init__(self, *args, **options):
		if not isinstance(self.definition, list) or not self.action in [True,False] or self.id == None:
			raise ProTRGError('Internal','Invalid condition/action definition (action = %s, definition = %s, id = %s, name = %s)' % (self.action,self.definition,self.id,self.name))
		self.data = [0] * [8,10][self.action]
		if len(args) != len(self.definition):
			raise ProTRGError('Parameters','Incorrect parameter count. Expected %s parameters but got %s. Definition:\n%s' % (len(self.definition),len(args),self.getDefinition()))
		for a,d in zip(args,self.definition):
			self.data[d[1]] = d[0].compile(a)
		add = options.get('add',  True)
		if 'add' in options:
			del options['add']
		if 'disabled' in options:
			self.disabled = not not options['disabled']
			del options['disabled']
		if options:
			raise ProTRGError('Interpreting','There are ony two valid keyword arguments, "add" and "disabled"')
		if add:
			t = None
			if add == True:
				if not TRIGGERS.lastTrigger:
					raise ProTRGError('Interpreting','You are attempting to add a%s when there are no triggers defined.' % [' condition','n action'][self.action])
				t = TRIGGERS.lastTrigger
			elif isinstance(add, Trigger):
				t = add
			else:
				raise
			t.add(self)

	def __getitem__(p):
		if p < 0 or p >= len(self.definition):
			raise
		return self.data[self.definition[p][1]]
	getParam = __getitem__

	def __setitem__(p, v):
		if p < 0 or p >= len(self.definition):
			raise
		self.data[self.definition[p][1]] = self.definition[p][0].compile(v)
	setParam = __setitem__

	def getDefinition(self):
		done = []
		p = []
		for d in self.definition:
			if not d[0] in done:
				p.append('%s: %s' % (d[0].__name__,d[0].__doc__))
				done.append(d[0])
		return '\n%s(%s)\n\t%s\n\n%s' % (self.__class__.__name__,", ".join([d[0].__name__ for d in self.definition]),self.__doc__,'\n'.join(p))

	def __str__(self):
		return "\t%s(%s);" % ([self.__class__.__name__,self.name][isstr(self.name)],", ".join([d[0].decompile(self.data[d[1]]) for d in self.definition]))

	def TRIG(self):
		self.data[[5,7][self.action]] = abs(self.id)
		p = 0
		for t,n in self.definition:
			if not TRIGGERS.trig == 2 and self.action:
				if t == PROPERTY:
					if not self.data[n] in TRIGGERS.added:
						p = 1
						#print self.data[n],'Property'
						#print TRIGGERS.properties,self.data[n],TRIGGERS.properties[self.data[n]]
						TRIGGERS.next.append(TRIGGERS.properties[self.data[n]].TRIG())
						TRIGGERS.added.append(self.data[n])
				elif t == STRING and self.data[n] != 0:
					if not -self.data[n] in TRIGGERS.added:
						s = TRIGGERS.strings[self.data[n]-1]
						#print self.data[n],s
						TRIGGERS.next.append(s + '\x00'*(2048-len(s)))
						TRIGGERS.added.append(-self.data[n])
			self.data[n] = int(self.data[n])
		if self.id > 0:
			if self.action:
				self.data[9] = 2*self.disabled + 4*self.data[9] + 8*p
			else:
				self.data[7] = 2*self.disabled
		#print ['','\t'][self.action],self.definition
		#print ['','\t'][self.action],tuple(self.data)
		if self.action:
			return struct.pack('<6LH3B3x', *self.data)
		return struct.pack('<3LH4Bxx', *self.data)

class Condition(Raw):
	action = False

class Action(Raw):
	action = True

class Property:
	valid_properties = ["Cloaked","Burrowed","InTransit","Hallucinated","Invincible"]
	valid_data = {"Health":(PERCENT,0),"Shields":(PERCENT,1),"Energy":(PERCENT,2),"Resources":(LONG,3),"HangerCount":(SHORT,4)}
	def __init__(self, *properties, **data):
		if len(TRIGGERS.properties) == 64:
			raise ProTRGError('Property','You can only have 64 different Property\'s')
		TRIGGERS.properties.append(self)
		self.properties = [0]*5
		for p in properties:
			if not p in self.valid_properties:
				raise ProTRGError('Property',""""%s" is an invalid special property. Valid special properties are:
	Cloaked = "Cloaked"
	Burrowed = "Burrowed"
	InTransit = "InTransit"
	Hallucinated = "Hallucinated"
	Invincible = "Invincible"
(got %s)""" % p)
			else:
				self.properties[self.valid_properties.index(p)] = 1
		self.data = [0]*5
		for n,d in data.iteritems():
			if not n in self.valid_data:
				raise ProTRGError('Property',""""%s" is an invalid Property data field. Valid data fields are:
	Health (takes a PERCENT)
	Shields (takes a PERCENT)
	Energy (takes a PERCENT)
	Resources (takes a LONG)
	HangerCount (takes a SHORT)
(got %s)""" % n)
			else:
				t,p = self.valid_data[n]
				self.properties[p] = t.compile(d)

	def TRIG(self):
		flags = sum((n*n*v for n,v in zip((1,2,4,8,16),self.properties)))
		return struct.pack('<HHx3BLHH4x', 31, 31, *(self.data+[flags]))

# Conditions
class CountdownTimer(Condition):
	"""Countdown timer is COMPARISON LONG game seconds."""
	id = 1
	definition = [(COMPARISON,4),(LONG,2)]
	name = "Countdown Timer"

class Command(Condition):
	"""PLAYER commands COMPARISON LONG TUNIT."""
	id = 2
	definition = [(PLAYER,1),(TUNIT,3),(COMPARISON,4),(LONG,2)]

class Bring(Condition):
	"""PLAYER brings COMPARISON LONG TUNIT to LOCATION."""
	id = 3
	definition = [(PLAYER,1),(TUNIT,3),(LOCATION,0),(COMPARISON,4),(LONG,2)]

class Accumulate(Condition):
	"""PLAYER accumulates COMPARISON LONG RESOURCE."""
	id = 4
	definition = [(PLAYER,1),(COMPARISON,4),(LONG,2),(RESOURCE,6)]

class Kill(Condition):
	"""PLAYER kills COMPARISON LONG TUNIT."""
	id = 5
	definition = [(PLAYER,1),(TUNIT,3),(COMPARISON,4),(LONG,2)]

class CommandTheMost(Condition):
	"""Current player commands the most TUNIT."""
	id = 6
	definition = [(TUNIT,3)]
	name = "Command the Most"

class CommandTheMostAt(Condition):
	"""Current player commands the most TUNIT at LOCATION."""
	id = 7
	definition = [(TUNIT,3),(LOCATION,0)]
	name = "Commands the Most At"

class MostKills(Condition):
	"""Current player has most kills of TUNIT."""
	id = 8
	definition = [(TUNIT,3)]
	name = "Most Kills"

class HighestScore(Condition):
	"""Current player has highest SCORE."""
	id = 9
	definition = [(SCORE,6)]
	name = "Highest Score"

class MostResources(Condition):
	"""Current player has most RESOURCE."""
	id = 10
	definition = [(RESOURCE,6)]
	name = "Most Resources"

class Switch(Condition):
	"""SWITCH is STATE."""
	id = 11
	definition = [(SWITCH,6),(STATE,4)]

class ElapsedTime(Condition):
	"""Elapsed scenario time is COMPARISON LONG game seconds."""
	id = 12
	definition = [(COMPARISON,4),(LONG,2)]
	name = "Elapsed Time"

class Opponents(Condition):
	"""PLAYER has COMPARISON LONG opponents remaining in the game."""
	id = 14
	definition = [(PLAYER,1),(COMPARISON,4),(LONG,2)]

class Deaths(Condition):
	"""PLAYER has suffered COMPARISON LONG deaths of EXTENDEDUNIT."""
	id = 15
	definition = [(PLAYER,1),(EXTENDEDUNIT,3),(COMPARISON,4),(LONG,2)]

class CommandTheLeast(Condition):
	"""Current player commands the least TUNIT."""
	id = 16
	definition = [(TUNIT,3)]
	name = "Command the Least"

class CommandTheLeastAt(Condition):
	"""Current player commands the least TUNIT at LOCATION."""
	id = 17
	definition = [(TUNIT,3),(LOCATION,0)]
	name = "Command the Least At"

class LeastKills(Condition):
	"""Current player has least kills of TUNIT."""
	id = 18
	definition = [(TUNIT,3)]
	name = "Least Kills"

class LowestScore(Condition):
	"""Current player has lowest SCORE."""
	id = 19
	definition = [(SCORE,6)]
	name = "Lowest Score"

class LeastResources(Condition):
	"""Current player has least RESOURCE."""
	id = 20
	definition = [(RESOURCE,6)]
	name = "Least Resources"

class Score(Condition):
	"""PLAYER SCORE score is COMPARISON LONG."""
	id = 21
	definition = [(PLAYER,1),(SCORE,6),(COMPARISON,4),(LONG,2)]

class Always(Condition):
	"""Always."""
	id = 22
	definition = []

class Never(Condition):
	"""Never."""
	id = 23
	definition = []

class Memory(Condition):
	"""Never."""
	id = 23
	definition = [(LONG,1),(COMPARISON,4),(LONG,2)]

# Actions
class Victory(Action):
	"""End scenario in victory for current player."""
	id = 1
	definition = []

class Defeat(Action):
	"""End scenario in defeat for current player."""
	id = 2
	definition = []

class PreserveTrigger(Action):
	"""Preserve trigger."""
	id = 3
	definition = []
	name = "Preserve Trigger"

class Wait(Action):
	"""Wait for LONG milliseconds."""
	id = 4
	definition = [(LONG,3)]

class PauseGame(Action):
	"""Pause the game."""
	id = 5
	definition = []
	name = "Pause Game"

class UnpauseGame(Action):
	"""Unpause the game."""
	id = 6
	definition = []
	name = "Unpause Game"

class Transmission(Action):
	"""Send transmission to current player from UNIT at LOCATION. Play wav STRING(2) for LONG(2) milliseconds(?). Modify transmission duration: MODIFIER LONG milliseconds. Display text message STRING with display property SUBTITLE."""
	id = 7
	definition = [(SUBTITLE,9),(STRING,1),(UNIT,6),(LOCATION,0),(MODIFIER,8),(LONG,5),(STRING,2),(LONG,3)]

class PlayWAV(Action):
	"""Play WAV for LONG milliseconds(?)."""
	id = 8
	definition = [(STRING,2),(LONG,3)]
	name = "Play WAV"

class DisplayTextMessage(Action):
	"""Display STRING for current player with display property SUBTITLE."""
	id = 9
	definition = [(SUBTITLE,9),(STRING,1)]
	name = "Display Text Message"

class CenterView(Action):
	"""Center view for current player at LOCATION."""
	id = 10
	definition = [(LOCATION,0)]
	name = "Center View"

class CreateUnitWithProperties(Action):
	"""Create LONG UNIT at LOCATION for PLAYER. Apply PROPERTY."""
	id = 11
	definition = [(PLAYER,4),(UNIT,6),(LONG,8),(LOCATION,0),(PROPERTY,5)]
	name = "Create Unit with Properties"

class SetMissionObjectives(Action):
	"""Set mission objectives to STRING."""
	id = 12
	definition = [(STRING,1)]
	name = "Set Mission Objectives"

class SetSwitch(Action):
	"""SETSWITCHSTATE on SWITCH."""
	id = 13
	definition = [(SWITCH,5),(SETSWITCHSTATE,8)]
	name = "Set Switch"

class SetCountdownTimer(Action):
	"""Modify countdown timer: MODIFIER LONG seconds."""
	id = 14
	definition = [(MODIFIER,8),(LONG,3)]
	name = "Set Countdown Timer"

class RunAIScript(Action):
	"""Execute AI Script AISCRIPT."""
	id = 15
	definition = [(AISCRIPT,5)]
	name = "Run AI Script"

class RunAIScriptAtLocation(Action):
	"""Execute AI Script AISCRIPT at LOCATION."""
	id = 16
	definition = [(AISCRIPT,5),(LOCATION,0)]
	name = "Run AI Script At Location"

class LeaderBoardControl(Action):
	"""Show Leader Board for most control of TUNIT. Display label STRING."""
	id = 17
	definition = [(STRING,1),(TUNIT,6)]
	name = "Leader Board Control"

class LeaderBoardControlAtLocation(Action):
	"""Show Leader Board for most control of TUNIT at LOCATION. Display label STRING."""
	id = 18
	definition = [(STRING,1),(TUNIT,6),(LOCATION,0)]
	name = "Leader Board Control At Location"

class LeaderBoardResources(Action):
	"""Show Leader Board for accumulation of most RESOURCE. Display label STRING."""
	id = 19
	definition = [(STRING,1),(RESOURCE,6)]
	name = "Leader Board Resources"

class LeaderBoardKills(Action):
	"""Show Leader Board for player closest to LONG kills of TUNIT. Display label STRING."""
	id = 20
	definition = [(STRING,1),(TUNIT,6)]
	name = "Leader Board Kills"

class LeaderBoardPoints(Action):
	"""Show Leader Board for player closest to LONG of SCORE. Display label STRING."""
	id = 21
	definition = [(STRING,1),(SCORE,6)]
	name = "Leader Board Points"

class KillUnit(Action):
	"""Kill all TUNIT for PLAYER."""
	id = 22
	definition = [(PLAYER,4),(TUNIT,6)]
	name = "Kill Unit"

class KillUnitAtLocation(Action):
	"""Kill QUANTITY TUNIT for PLAYER at LOCATION."""
	id = 23
	definition = [(PLAYER,4),(TUNIT,6),(QUANTITY,8),(LOCATION,0)]
	name = "Kill Unit At Location"

class RemoveUnit(Action):
	"""Remove all TUNIT for PLAYER."""
	id = 24
	definition = [(PLAYER,4),(TUNIT,6)]
	name = "Remove Unit"

class RemoveUnitAtLocation(Action):
	"""Remove QUANTITY TUNIT for PLAYER at LOCATION."""
	id = 25
	definition = [(PLAYER,4),(TUNIT,6),(QUANTITY,8),(LOCATION,0)]
	name = "Remove Unit At Location"

class SetResources(Action):
	"""Modify resources for PLAYER: MODIFIER LONG of RESOURCE."""
	id = 26
	definition = [(PLAYER,4),(MODIFIER,8),(LONG,5),(RESOURCE,6)]
	name = "Set Resources"

class SetScore(Action):
	"""Modify score for PLAYER: MODIFIER LONG of SCORE."""
	id = 27
	definition = [(PLAYER,4),(MODIFIER,8),(LONG,5),(SCORE,6)]
	name = "Set Score"

class MinimapPing(Action):
	"""Show minimap ping for current player at LOCATION."""
	id = 28
	definition = [(LOCATION,0)]
	name = "Minimap Ping"

class TalkingPortrait(Action):
	"""Show UNIT talking to current player for LONG milliseconds."""
	id = 29
	definition = [(UNIT,6),(LONG,3)]
	name = "Talking Portrait"

class MuteUnitSpeech(Action):
	"""Mute all non-trigger unit sounds for current player."""
	id = 30
	definition = []
	name = "Mute Unit Speech"

class UnmuteUnitSpeech(Action):
	"""Unmute all non-trigger unit sounds for current player."""
	id = 31
	definition = []
	name = "Unmute Unit Speech"

class LeaderboardComputerPlayers(Action):
	"""Set use of computer players in leaderboard calculations to STATE."""
	id = 32
	definition = [(SETOTHERSTATE,8)]
	name = "Leaderboard Computer Players"

class LeaderboardGoalControl(Action):
	"""Show Leader Board for player closest to control of LONG of TUNIT. Display label STRING."""
	id = 33
	definition = [(STRING,1),(TUNIT,6),(LONG,5)]
	name = "Leaderboard Goal Control"

class LeaderboardGoalControlAtLocation(Action):
	"""Show Leader Board for player closest to control of LONG of TUNIT at LOCATION. Display label STRING."""
	id = 34
	definition = [(STRING,1),(TUNIT,6),(LONG,5),(LOCATION,0)]
	name = "Leaderboard Goal Control At Location"

class LeaderboardGoalResources(Action):
	"""Show Leader Board for player closest to accumulation of LONG RESOURCE. Display label STRING"""
	id = 35
	definition = [(STRING,1),(LONG,5),(RESOURCE,6)]
	name = "Leaderboard Goal Resources"

class LeaderboardGoalKills(Action):
	"""Show Leader Board for player closest to LONG kills of TUNIT. Display label STRING."""
	id = 36
	definition = [(STRING,1),(TUNIT,6),(LONG,5)]
	name = "Leaderboard Goal Kills"

class LeaderboardGoalPoints(Action):
	"""Show Leader Board for player closest to LONG of SCORE. Display label STRING."""
	id = 37
	definition = [(STRING,1),(SCORE,6),(LONG,5)]
	name = "Leaderboard Goal Points"

class MoveLocation(Action):
	"""Center location LOCATION on TUNIT owned by PLAYER at LOCATION(2)."""
	id = 38
	definition = [(PLAYER,4),(TUNIT,6),(LOCATION,0),(LOCATION,5)]
	name = "Move Location"

class MoveUnit(Action):
	"""Move QUANTITY TUNIT for PLAYER at LOCATION to LOCATION(2)."""
	id = 39
	definition = [(PLAYER,4),(TUNIT,6),(QUANTITY,8),(LOCATION,0),(LOCATION,5)]
	name = "Move Unit"

class LeaderboardGreed(Action):
	"""Show Greed Leader Board for player closest to accumulation of LONG ore and gas."""
	id = 40
	definition = [(LONG,5)]
	name = "Leaderboard Greed"

class SetNextScenario(Action):
	"""Load scenario STRING after completion of current game."""
	id = 41
	definition = [(STRING,1)]
	name = "Set Next Scenario"

class SetDoodadState(Action):
	"""Set doodad state for TUNIT for PLAYER at LOCATION to SETOTHERSTATE."""
	id = 42
	definition = [(PLAYER,4),(TUNIT,6),(LOCATION,0),(SETOTHERSTATE,8)]
	name = "Set Doodad State"

class SetInvincibility(Action):
	"""Set invincibility for TUNIT owned by PLAYER at LOCATION to SETOTHERSTATE."""
	id = 43
	definition = [(PLAYER,4),(TUNIT,6),(LOCATION,0),(SETOTHERSTATE,8)]
	name = "Set Invincibility"

class CreateUnit(Action):
	"""Create LONG UNIT at LOCATION for PLAYER."""
	id = 44
	definition = [(PLAYER,4),(UNIT,6),(LONG,8),(LOCATION,0)]
	name = "Create Unit"

class SetDeaths(Action):
	"""Modify death counts for PLAYER: MODIFIER LONG for UNIT."""
	id = 45
	definition = [(PLAYER,4),(UNIT,6),(MODIFIER,8),(LONG,5)]
	name = "Set Deaths"

class Order(Action):
	"""Issue order to all TUNIT owned by PLAYER at LOCATION: Order to LOCATION(2)."""
	id = 46
	definition = [(PLAYER,4),(TUNIT,6),(LOCATION,0),(LOCATION,5),(ORDER,8)]

class Comment(Action):
	"""Comment: STRING."""
	id = 47
	definition = [(STRING,1)]

class GiveUnitsToPlayer(Action):
	"""Give QUANTITY TUNIT owned by PLAYER at LOCATION to PLAYER(2)."""
	id = 48
	definition = [(PLAYER,4),(PLAYER,5),(TUNIT,6),(QUANTITY,8),(LOCATION,0)]
	name = "Give Units to Player"

class ModifyUnitHitPoints(Action):
	"""Set hit points for QUANTITY TUNIT owned by PLAYER at LOCATION to PERCENT."""
	id = 49
	definition = [(PLAYER,4),(TUNIT,6),(PERCENT,5),(QUANTITY,8),(LOCATION,0)]
	name = "Modify Unit Hit Points"

class ModifyUnitEnergy(Action):
	"""Set energy points for QUANTITY TUNIT owned by PLAYER at LOCATION to PERCENT."""
	id = 50
	definition = [(PLAYER,4),(TUNIT,6),(PERCENT,5),(QUANTITY,8),(LOCATION,0)]
	name = "Modify Unit Energy"

class ModifyUnitShieldPoints(Action):
	"""Set shield points for QUANTITY TUNIT owned by PLAYER at LOCATION to PERCENT"""
	id = 51
	definition = [(PLAYER,4),(TUNIT,6),(PERCENT,5),(QUANTITY,8),(LOCATION,0)]
	name = "Modify Unit Shield Points"

class ModifyUnitResourceAmount(Action):
	"""Set resource amount for QUANTITY resource sources owned by PLAYER at LOCATION to LONG."""
	id = 52
	definition = [(PLAYER,4),(LONG,5),(QUANTITY,8),(LOCATION,0)]
	name = "Modify Unit Resource Amount"

class ModifyUnitHangerCount(Action):
	"""Add at most LONG to hangar for QUANTITY TUNIT at LOCATION owned by PLAYER."""
	id = 53
	definition = [(PLAYER,4),(TUNIT,6),(LONG,5),(QUANTITY,8),(LOCATION,0)]
	name = "Modify Unit Hanger Count"

class PauseTimer(Action):
	"""Pause the countdown timer."""
	id = 54
	definition = []
	name = "Pause Timer"

class UnpauseTimer(Action):
	"""Unpause the countdown timer."""
	id = 55
	definition = []
	name = "Unpause Timer"

class Draw(Action):
	"""End the scenario in a draw for all players."""
	id = 56
	definition = []

class SetAllianceStatus(Action):
	"""Set PLAYER to AllyStatus."""
	id = 57
	definition = [(PLAYER,4),(ALLYSTATUS,6)]
	name = "Set Alliance Status"

# Not supported by SCMDraft:
class NoCondition(Condition):
	"""No condition. Not supported by SCMDraft."""
	id = 0
	definition = []

class RawCondition(Condition):
	"""Create a Raw condition. Not supported by SCMDraft."""
	id = 0
	definition = [(LONG,0),(LONG,1),(LONG,2),(SHORT,3),(BYTE,4),(BYTE,5),(BYTE,6),(BYTE,7)]

class NoAction(Action):
	"""No action.  Not supported by SCMDraft."""
	id = 0
	definition = []

class DisableDebugMode(Action):
	"""Disable debug mode (does nothing?). Not supported by SCMDraft."""
	id = -58
	definition = []

class EnableDebugMode(Action):
	"""Enable debug mode (does nothing?). Not supported by SCMDraft."""
	id = -59
	definition = []

class RawAction(Action):
	"""Create a Raw action. Not supported by SCMDraft."""
	id = 0
	definition = [(LONG,0),(LONG,1),(LONG,2),(LONG,3),(LONG,4),(LONG,5),(SHORT,6),(BYTE,7),(BYTE,8),(BYTE,9)]

# Main
def main():
	p = optparse.OptionParser(usage='usage: %s%s [options] [in [out]]' % (('python ','')[hasattr(sys, 'frozen')],protrg), version='ProTRG %s' % LONG_VERSION)
	p.add_option('-t', '--trg', action='store_true', dest='trg', help="Compile to a TRG file", default=False)
	p.add_option('-g', '--got', action='store_true', dest='got', help="Used with the --trg option to generate a GOT compatable TRG file", default=False)
	opt, args = p.parse_args()
	if not args:
		args.append('triggers%sprotrg' % os.extsep)
	SCRIPT_PATH = os.path.dirname(args[0])
	if len(args) == 1:
		args.append(os.path.join(SCRIPT_PATH,'%s%s%s%s' % (('scmd_','')[opt.trg], os.path.basename(args[0]).split(os.extsep)[0], os.extsep, ('txt','trg')[opt.trg])))
	INCLUDED = []
	include_re = re.compile("^\s*@include(?:\s+(.+?)?\s*)?$", re.MULTILINE)
	def include(m):
		if m.group(1) == None:
			raise SyntaxError("invalid syntax")
		if not m.group(1) in INCLUDED:
			for path in (SCRIPT_PATH,PROTRG_PATH):
				for ext in (None,'protrg','txt','py','pyw'):
					n = os.path.join(path,'%s%s%s' % (m.group(1), [os.extsep,''][ext == None], ext))
					if os.path.exists(n):
						INCLUDED.append(m.group(1))
						try:
							t = open(n, 'r')
							r = include_re.sub(include, t.read())
							t.close()
							return r
						except:
							pass
		raise ProTRGError('Include','Couldn\'t find "%s"' % m.group(1))
	try:
		f = open(args[0],'r')
		script = f.read()
		f.close()
	except:
		sys.stderr.write("Could not open file '%s'." % args[0])
	else:
		if opt.trg:
			sys.stdout = sys.stderr
			TRIGGERS.trig = 1 + opt.got
		else:
			sys.stderr.close()
			sys.stdout = sys.stderr = open(args[1],'w')
			m = " Generated by ProTRG %s on %s " % (LONG_VERSION,time.asctime())
			print '//%s\\\\\n//%s\\\\\n//%s\\\\\n\n' % ('-' * len(m),m,'-' * len(m))
		try:
			exec include_re.sub(include, script) in globals(), locals()
		except Exception, e:
			t = [list(t) for t in traceback.extract_tb(sys.exc_info()[2])]
			del t[0]
			if isinstance(e, ProTRGError):
				t[0][2] = '<triggers>'
				while os.path.basename(t[-1][0]).lower().startswith('protrg'):
					del t[-1]
			print "Traceback (most recent call last):"
			for l in traceback.format_list(t):
				print l,
			if isinstance(e, ProTRGError):
				print e
			else:
				for l in traceback.format_exception_only(type(e), e):
					print l,
			if TRIGGERS.triggerList:
				m = " Triggers generated before error occurred: "
				print '\n\n\n//%s\\\\\n//%s\\\\\n//%s\\\\\n\n' % ('-' * len(m),m,'-' * len(m))
				try:
					for z in TRIGGERS.triggerList:
						print z
				except:
					pass
		else:
			f.close()
			if opt.trg:
				trg = open(args[1], 'wb')
				if not opt.got:
					trg.write('qw\x986\x18\x00\x00\x00')
				for t in TRIGGERS.triggerList:
					c = t.TRIG()
					#print len(c)
					trg.write(c)
					if TRIGGERS.next:
						for n in TRIGGERS.next:
							#print '\t',len(n),n[:n.index('\x00')]
							trg.write(n)
						TRIGGERS.next = []
				trg.close()
			else:
				for t in TRIGGERS.triggerList:
					print t
		sys.stdout.close()

if __name__ == '__main__':
	main()
